@@ -10,7 +10,7 @@
    to maintain the source code under the licensing terms described
    herein and below.
 
-   Copyright © 2021 Logan Rathbone <poprocks@gmail.com>
+   Copyright © 2021-2022 Logan Rathbone <poprocks@gmail.com>
 
    GHex is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -125,30 +125,34 @@ inbtn_clicked_cb (GtkButton *button, gpointer user_data)
 
 	g_return_if_fail (GTK_IS_TREE_VIEW(treeview));
 
-	(void)button;	/* unused */
-
 	model = gtk_tree_view_get_model (treeview);
 
 	insert_char (treeview, model);
 }
 
-static void
-hide_chartable_cb (GtkButton *button, gpointer user_data)
+static gboolean
+key_press_cb (GtkEventControllerKey *key,
+  guint keyval,
+  guint keycode,
+  GdkModifierType state,
+  GtkWindow *ct)
 {
-	GtkWindow *win = GTK_WINDOW(user_data);
-
-	(void)button;	/* unused */
-
-	gtk_window_close (win);
+	if (keyval == GDK_KEY_Escape)
+	{
+		gtk_window_close (ct);
+		return GDK_EVENT_STOP;
+	}
+	return GDK_EVENT_PROPAGATE;
 }
 
-GtkWidget *create_char_table(GtkWindow *parent_win, HexWidget *gh)
+
+GtkWidget *create_char_table (GtkWindow *parent_win, HexWidget *gh)
 {
 	static gchar *fmt[] = { NULL, "%02X", "%03d", "%03o" };
 	static gchar *titles[] = {  N_("ASCII"), N_("Hex"), N_("Decimal"),
 		N_("Octal"), N_("Binary") };
 	gchar *real_titles[5];
-	GtkWidget *ct, *sw, *ctv, *inbtn, *cbtn, *vbox, *hbox, *lbl;
+	GtkWidget *ct, *sw, *ctv, *inbtn, *cbtn, *vbox, *hbox, *lbl, *sep;
 	GtkListStore *store;
 	GtkCellRenderer *cell_renderer;
 	GtkTreeViewColumn *column;
@@ -156,6 +160,7 @@ GtkWidget *create_char_table(GtkWindow *parent_win, HexWidget *gh)
 	GtkTreeSelection *selection;
 	int i, col;
 	gchar *label, ascii_printable_label[2], bin_label[9], *row[5];
+	GtkEventController *key;
 
 	/* set global HexWidget widget */
 	g_assert (HEX_IS_WIDGET(gh));
@@ -166,6 +171,10 @@ GtkWidget *create_char_table(GtkWindow *parent_win, HexWidget *gh)
 	 */
 	ct = gtk_window_new ();
 
+	key = gtk_event_controller_key_new ();
+	gtk_widget_add_controller (ct, key);
+	g_signal_connect (key, "key-pressed", G_CALLBACK(key_press_cb), ct);
+
 	if (parent_win) {
 		g_assert (GTK_IS_WINDOW (parent_win));
 
@@ -245,23 +254,27 @@ GtkWidget *create_char_table(GtkWindow *parent_win, HexWidget *gh)
 					G_CALLBACK(inbtn_clicked_cb), ctv);
 
 	cbtn = gtk_button_new_with_mnemonic (_("_Close"));
-	g_signal_connect(G_OBJECT (cbtn), "clicked",
-					G_CALLBACK(hide_chartable_cb), ct);
+	g_signal_connect_swapped (cbtn, "clicked", G_CALLBACK(gtk_window_close), ct);
 
 	lbl = gtk_label_new ("");
 	vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 4);
 	hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 4);
+	gtk_widget_set_name (hbox, "chartable-action-area");
+
+	sep = gtk_separator_new (GTK_ORIENTATION_HORIZONTAL);
 
-	gtk_box_append (GTK_BOX(vbox), sw);
 	gtk_box_append (GTK_BOX(hbox), lbl);
 	gtk_box_append (GTK_BOX(hbox), inbtn);
 	gtk_box_append (GTK_BOX(hbox), cbtn);
+
+	gtk_box_append (GTK_BOX(vbox), sw);
+	gtk_box_append (GTK_BOX(vbox), sep);
 	gtk_box_append (GTK_BOX(vbox), hbox);
 
 	gtk_scrolled_window_set_child (GTK_SCROLLED_WINDOW(sw), ctv);
-	gtk_window_set_child (GTK_WINDOW (ct), vbox);
+	gtk_window_set_child (GTK_WINDOW(ct), vbox);
 
-	gtk_widget_set_size_request(ct, 320, 320);
+	gtk_widget_set_size_request (ct, 320, 320);
 
 	return ct;
 }
