@@ -67,16 +67,6 @@ typedef enum {
 	VIEW_ASCII
 } HexWidgetViewType;
 
-enum {
-	CURSOR_MOVED_SIGNAL,
-	DATA_CHANGED_SIGNAL,
-	CUT_CLIPBOARD_SIGNAL,
-	COPY_CLIPBOARD_SIGNAL,
-	PASTE_CLIPBOARD_SIGNAL,
-	DRAW_COMPLETE_SIGNAL,
-	LAST_SIGNAL
-};
-
 /* highlighting information.
  */
 typedef struct _HexWidget_Highlight HexWidget_Highlight;
@@ -104,6 +94,8 @@ struct _HexWidgetAutoHighlight
 	char *search_string;
 	int search_len;
 
+	HexSearchFlags search_flags;
+
 	gint64 view_min;
 	gint64 view_max;
 
@@ -121,12 +113,35 @@ hex_widget_autohighlight_copy (HexWidgetAutoHighlight *ahl)
 G_DEFINE_BOXED_TYPE (HexWidgetAutoHighlight, hex_widget_autohighlight,
 		hex_widget_autohighlight_copy, g_free)
 
-
 /* ------------------------------
  * Main HexWidget GObject definition
  * ------------------------------
  */
 
+/* PROPERTIES */
+
+enum
+{
+	DOCUMENT = 1,
+	N_PROPERTIES
+};
+
+static GParamSpec *properties[N_PROPERTIES];
+
+/* SIGNALS */
+
+enum {
+	CURSOR_MOVED_SIGNAL,
+	DATA_CHANGED_SIGNAL,
+	CUT_CLIPBOARD_SIGNAL,
+	COPY_CLIPBOARD_SIGNAL,
+	PASTE_CLIPBOARD_SIGNAL,
+	DRAW_COMPLETE_SIGNAL,
+	LAST_SIGNAL
+};
+
+static guint gtkhex_signals[LAST_SIGNAL] = { 0 };
+
 /**
  * HexWidget:
  *
@@ -202,15 +217,126 @@ struct _HexWidget
 	/* default characters per line and number of lines. */
 	int default_cpl;
 	int default_lines;
+
+	GdkContentProvider *selection_content;
 };
 
 G_DEFINE_TYPE (HexWidget, hex_widget, GTK_TYPE_WIDGET)
 
 /* ----- */
 
-/* STATIC FORWARD DECLARATIONS */
+/* HexContentProvider */
 
-static guint gtkhex_signals[LAST_SIGNAL] = { 0 };
+#define HEX_TYPE_CONTENT_PROVIDER (hex_content_provider_get_type ())
+G_DECLARE_FINAL_TYPE (HexContentProvider, hex_content_provider, HEX, CONTENT_PROVIDER,
+		GdkContentProvider)
+
+struct _HexContentProvider
+{
+	GdkContentProvider parent_instance;
+
+	HexWidget *owner;
+};
+
+G_DEFINE_TYPE (HexContentProvider, hex_content_provider, GDK_TYPE_CONTENT_PROVIDER)
+
+static GdkContentFormats *
+hex_content_provider_ref_formats (GdkContentProvider *provider)
+{
+	HexContentProvider *content = HEX_CONTENT_PROVIDER (provider);
+	HexWidget *self = content->owner;
+	GdkContentFormatsBuilder *builder = gdk_content_formats_builder_new ();
+
+	gdk_content_formats_builder_add_gtype (builder, HEX_TYPE_PASTE_DATA);
+	gdk_content_formats_builder_add_gtype (builder, G_TYPE_STRING);
+
+	return gdk_content_formats_builder_free_to_formats (builder);
+}
+
+static void
+hex_content_provider_detach (GdkContentProvider *provider,
+		GdkClipboard *clipboard)
+{
+	HexContentProvider *content = HEX_CONTENT_PROVIDER (provider);
+	HexWidget *self = content->owner;
+
+	self->selecting = FALSE;
+	hex_widget_set_selection (self, self->cursor_pos, self->cursor_pos);
+}
+
+static gboolean
+hex_content_provider_get_value (GdkContentProvider *provider,
+		GValue *value,
+		GError **error)
+{
+	HexContentProvider *content = HEX_CONTENT_PROVIDER (provider);
+	HexWidget *self = content->owner;
+	HexPasteData *paste;
+	gint64 start_pos, end_pos;
+	size_t len;
+	char *doc_data;
+
+	/* cross-ref: hex_widget_real_copy_to_clipboard - similar initial code */
+
+	start_pos = MIN(self->selection.start, self->selection.end);
+	end_pos = MAX(self->selection.start, self->selection.end);
+
+	len = end_pos - start_pos + 1;
+	g_return_val_if_fail (len, FALSE);
+
+	doc_data = hex_buffer_get_data (hex_document_get_buffer(self->document),
+			start_pos, len);
+
+	paste = hex_paste_data_new (doc_data, len);
+	g_return_val_if_fail (HEX_IS_PASTE_DATA(paste), FALSE);
+
+	if (G_VALUE_HOLDS (value, G_TYPE_STRING))
+	{
+		char *string;
+
+		string = hex_paste_data_get_string (paste);
+		g_value_take_string (value, string);
+		g_object_unref (paste);
+
+		return TRUE;
+	}
+	else if (G_VALUE_HOLDS (value, HEX_TYPE_PASTE_DATA))
+	{
+		g_value_take_object (value, paste);
+
+		return TRUE;
+	}
+
+	/* chain up */
+	return GDK_CONTENT_PROVIDER_CLASS (hex_content_provider_parent_class)->get_value (
+			provider, value, error);
+}
+
+static void
+hex_content_provider_class_init (HexContentProviderClass *klass)
+{
+	GdkContentProviderClass *provider_class = GDK_CONTENT_PROVIDER_CLASS(klass);
+
+	provider_class->ref_formats = hex_content_provider_ref_formats;
+	provider_class->get_value = hex_content_provider_get_value;
+	provider_class->detach_clipboard = hex_content_provider_detach;
+}
+
+static void
+hex_content_provider_init (HexContentProvider *content)
+{
+}
+
+GdkContentProvider *
+hex_content_provider_new (void)
+{
+	return g_object_new (HEX_TYPE_CONTENT_PROVIDER, NULL);
+}
+
+/* --- */
+
+
+/* STATIC FORWARD DECLARATIONS */
 
 static char *char_widths = NULL;
 
@@ -238,6 +364,51 @@ static void recalc_displays (HexWidget *self);
 
 static void show_cursor (HexWidget *self, gboolean show);
 
+
+/* PROPERTIES - GETTERS AND SETTERS */
+
+static void
+hex_widget_set_property (GObject *object,
+		guint property_id,
+		const GValue *value,
+		GParamSpec *pspec)
+{
+	HexWidget *self = HEX_WIDGET(object);
+
+	switch (property_id)
+	{
+		case DOCUMENT:
+			self->document = g_value_get_object (value);
+			break;
+
+		default:
+			/* We don't have any other property... */
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+			break;
+	}
+}
+
+static void
+hex_widget_get_property (GObject *object,
+		guint property_id,
+		GValue *value,
+		GParamSpec *pspec)
+{
+	HexWidget *self = HEX_WIDGET(object);
+
+	switch (property_id)
+	{
+		case DOCUMENT:
+			g_value_set_object (value, self->document);
+			break;
+
+		default:
+			/* We don't have any other property... */
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+			break;
+	}
+}
+
 /* HexWidget - Method Definitions */
 
 /* ACTIONS */
@@ -249,9 +420,6 @@ copy_action (GtkWidget *widget,
 {
 	HexWidget *self = HEX_WIDGET(widget);
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	(void)action_name, (void)parameter;
-
 	hex_widget_copy_to_clipboard (self);
 }
 
@@ -262,9 +430,6 @@ cut_action (GtkWidget *widget,
 {
 	HexWidget *self = HEX_WIDGET(widget);
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	(void)action_name, (void)parameter;
-
 	hex_widget_cut_to_clipboard (self);
 }
 
@@ -275,9 +440,6 @@ paste_action (GtkWidget *widget,
 {
 	HexWidget *self = HEX_WIDGET(widget);
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	(void)action_name, (void)parameter;
-
 	hex_widget_paste_from_clipboard (self);
 }
 
@@ -289,10 +451,7 @@ redo_action (GtkWidget *widget,
 	HexWidget *self = HEX_WIDGET(widget);
 	HexChangeData *cd;
 
-	(void)action_name, (void)parameter;
-
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	g_return_if_fail (HEX_IS_DOCUMENT(self->document));
+	g_return_if_fail (HEX_IS_DOCUMENT (self->document));
 
 	if (hex_document_can_redo (self->document))
 	{
@@ -309,7 +468,6 @@ static void
 doc_undo_redo_cb (HexDocument *doc, gpointer user_data)
 {
 	HexWidget *self = HEX_WIDGET(user_data);
-	g_return_if_fail (HEX_IS_WIDGET (self));
 	
 	gtk_widget_action_set_enabled (GTK_WIDGET(self),
 			"gtkhex.undo", hex_document_can_undo (doc));
@@ -325,10 +483,7 @@ undo_action (GtkWidget *widget,
 	HexWidget *self = HEX_WIDGET(widget);
 	HexChangeData *cd;
 
-	(void)action_name, (void)parameter;
-
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	g_return_if_fail (HEX_IS_DOCUMENT(self->document));
+	g_return_if_fail (HEX_IS_DOCUMENT (self->document));
 
 	if (hex_document_can_undo (self->document))
 	{
@@ -348,8 +503,11 @@ toggle_hex_action (GtkWidget *widget,
 {
 	HexWidget *self = HEX_WIDGET (widget);
 
-	self->active_view = VIEW_HEX;
-	gtk_widget_queue_draw (widget);
+	if (gtk_widget_get_visible (self->xdisp))
+	{
+		self->active_view = VIEW_HEX;
+		gtk_widget_queue_draw (widget);
+	}
 }
 
 static void
@@ -359,8 +517,11 @@ toggle_ascii_action (GtkWidget *widget,
 {
 	HexWidget *self = HEX_WIDGET (widget);
 
-	self->active_view = VIEW_ASCII;
-	gtk_widget_queue_draw (widget);
+	if (gtk_widget_get_visible (self->adisp))
+	{
+		self->active_view = VIEW_ASCII;
+		gtk_widget_queue_draw (widget);
+	}
 }
 
 /*
@@ -911,62 +1072,6 @@ end_of_loop:
 	}
 }
 
-/* FIXME - Previously, this function was more sophisticated, and only
- * redrew part of the drawing area requested. Need to make an executive
- * decision as to whether that will be feasible to do for ghex4, or just
- * eliminate those and queue a redraw for the drawing area in question.
- */
-/*
- * when calling invalidate_*_lines() the imin and imax arguments are the
- * numbers of the first and last line TO BE INVALIDATED in the range
- * [0 .. self->vis_lines-1] AND NOT [0 .. self->lines]!
- */
-static void
-invalidate_lines (HexWidget *self,
-                  GtkWidget *widget,
-                  int imin,
-                  int imax)
-{
-#if 0
-    gtk_widget_get_allocation (widget, &allocation);
-#endif
-
-	(void)self, (void)imin, (void)imax; /* unused for now. See comment above. */
-
-    gtk_widget_queue_draw (widget);
-#if 0
-    gtk_widget_queue_draw_area (widget,
-                                0,
-                                imin * self->char_height,
-                                allocation.width,
-                                (imax - imin + 1) * self->char_height);
-#endif
-}
-
-static void
-invalidate_hex_lines (HexWidget *self,
-                      int imin,		/* FIXME - imin/imax for next few funcs */
-                      int imax)		/* presently ignored. */
-{
-    invalidate_lines (self, self->xdisp, imin, imax);
-}
-
-static void
-invalidate_ascii_lines (HexWidget *self,
-                        int imin,
-                        int imax)
-{
-    invalidate_lines (self, self->adisp, imin, imax);
-}
-
-static void
-invalidate_offsets (HexWidget *self,
-                    int imin,
-                    int imax)
-{
-    invalidate_lines (self, self->offsets, imin, imax);
-}
-
 /*
  * when calling render_lines() the min_lines and max_lines arguments are the
  * numbers of the first and last line TO BE DISPLAYED in the range
@@ -1153,15 +1258,20 @@ render_offsets (HexWidget *self,
 #undef MIN_CPL
 
 static void
-allocate_display_buffer (HexWidget *self)
+allocate_display_buffer (HexWidget *self, HexWidgetViewType type)
 {
 	GtkWidget *widget = GTK_WIDGET (self);
-	int hex_cpl = get_hex_cpl (self);
+	int cpl;
+
+	if (type == VIEW_HEX)
+		cpl = get_hex_cpl (self);
+	else
+		cpl = self->cpl;
 
 	if (self->disp_buffer)
 		g_free (self->disp_buffer);
 
-	self->disp_buffer = g_malloc ((hex_cpl + 1) * (self->vis_lines + 1));
+	self->disp_buffer = g_malloc ((cpl + 1) * (self->vis_lines + 1));
 }
 
 /* draw_func for the hex drawing area
@@ -1175,13 +1285,11 @@ hex_draw (GtkDrawingArea *drawing_area,
 {
 	HexWidget *self = HEX_WIDGET(user_data);
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-
 	/* Now that we have self->cpl defined, run this function to bump all
 	 * required values:
 	 */
 	recalc_displays (self);
-	allocate_display_buffer (self);
+	allocate_display_buffer (self, VIEW_HEX);
 
 	/* Finally, we can do what we wanted to do to begin with: draw our hex
 	 * lines!
@@ -1197,10 +1305,9 @@ ascii_draw (GtkDrawingArea *drawing_area,
                            gpointer user_data)
 {
 	HexWidget *self = HEX_WIDGET(user_data);
-	g_return_if_fail(HEX_IS_WIDGET(self));
 
 	recalc_displays (self);
-	allocate_display_buffer (self);
+	allocate_display_buffer (self, VIEW_ASCII);
 	render_lines (self, cr, 0, self->vis_lines, VIEW_ASCII);
 }
 
@@ -1212,7 +1319,6 @@ offsets_draw (GtkDrawingArea *drawing_area,
                            gpointer user_data)
 {
 	HexWidget *self = HEX_WIDGET(user_data);
-	g_return_if_fail(HEX_IS_WIDGET(self));
 
 	recalc_displays (self);
 	render_offsets (self, cr, 0, self->vis_lines);
@@ -1293,8 +1399,9 @@ adj_value_changed_cb (GtkAdjustment *adj, HexWidget *self)
 {
 	int dx, dy;
 	
-	g_return_if_fail (gtk_widget_is_drawable (self->xdisp) &&
-			gtk_widget_is_drawable (self->adisp));
+	if (! gtk_widget_is_drawable (self->xdisp) &&
+			! gtk_widget_is_drawable (self->adisp))
+		return;
 
 	self->top_line = gtk_adjustment_get_value (adj);
 
@@ -1330,8 +1437,6 @@ scroll_cb (GtkEventControllerScroll *controller,
 	HexWidget *self = HEX_WIDGET (user_data);
 	double old_value, new_value;
 
-	g_return_val_if_fail (HEX_IS_WIDGET(self), FALSE);
-
 	old_value = gtk_adjustment_get_value (self->adj);
 	new_value = old_value + dy;
 
@@ -1359,9 +1464,6 @@ pressed_gesture_helper (HexWidget *self,
 	else
 		widget = self->adisp;
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	g_return_if_fail (GTK_IS_WIDGET(widget));
-
 	button = gtk_gesture_single_get_current_button (GTK_GESTURE_SINGLE(gesture));
 
 	/* Single-press */
@@ -1399,6 +1501,110 @@ pressed_gesture_helper (HexWidget *self,
 	}
 }
 
+static void
+update_primary_selection (HexWidget *self)
+{
+	GtkWidget *widget = GTK_WIDGET(self);
+	GdkClipboard *clipboard;
+
+	if (! gtk_widget_get_realized (widget))
+		return;
+
+	clipboard = gtk_widget_get_primary_clipboard (widget);
+
+	if (self->selection.start != self->selection.end)
+	{
+		gdk_clipboard_set_content (clipboard, self->selection_content);
+	}
+	else
+	{
+		if (gdk_clipboard_get_content (clipboard) == self->selection_content)
+			gdk_clipboard_set_content (clipboard, NULL);
+	}
+}
+
+static void
+plaintext_paste_received_cb (GObject *source_object,
+		GAsyncResult *result,
+		gpointer user_data)
+{
+	HexWidget *self = HEX_WIDGET(user_data);
+	GdkClipboard *clipboard;
+	char *text;
+	GError *error = NULL;
+
+	g_debug ("%s: We DON'T have HexPasteData. Falling back to plaintext paste",
+			__func__);
+
+	clipboard = GDK_CLIPBOARD (source_object);
+
+	/* Get the resulting text of the read operation */
+	text = gdk_clipboard_read_text_finish (clipboard, result, &error);
+
+	if (text) {
+		hex_document_set_data (self->document,
+				self->cursor_pos,
+				strlen(text),
+				0,	/* rep_len (0 to insert w/o replacing; what we want) */
+				text,
+				TRUE);
+
+		hex_widget_set_cursor (self, self->cursor_pos + strlen(text));
+		g_free(text);
+	}
+	else {
+		g_critical ("Error pasting text: %s",
+				error->message);
+		g_error_free (error);
+	}
+}
+
+static void
+paste_helper (HexWidget *self, GdkClipboard *clipboard)
+{
+	GdkContentProvider *content;
+	GValue value = G_VALUE_INIT;
+	HexPasteData *paste;
+	gboolean have_hex_paste_data = FALSE;
+
+	content = gdk_clipboard_get_content (clipboard);
+	g_value_init (&value, HEX_TYPE_PASTE_DATA);
+
+	/* If the clipboard contains our special HexPasteData, we'll use it.
+	 * If not, just fall back to plaintext.
+	 */
+	have_hex_paste_data = content ?
+		gdk_content_provider_get_value (content, &value, NULL) : FALSE;
+
+	if (have_hex_paste_data)
+	{
+		char *doc_data;
+		int elems;
+
+		g_debug("%s: We HAVE HexPasteData.", __func__);
+
+		paste = HEX_PASTE_DATA(g_value_get_object (&value));
+		doc_data = hex_paste_data_get_doc_data (paste);
+		elems = hex_paste_data_get_elems (paste);
+
+		hex_document_set_data (self->document,
+				self->cursor_pos,
+				elems,
+				0,	/* rep_len (0 to insert w/o replacing; what we want) */
+				doc_data,
+				TRUE);
+
+		hex_widget_set_cursor (self, self->cursor_pos + elems);
+	}
+	else
+	{
+		gdk_clipboard_read_text_async (clipboard,
+				NULL,	/* cancellable */
+				plaintext_paste_received_cb,
+				self);
+	}
+}
+
 static void
 released_gesture_helper (HexWidget *self,
 		GtkGestureClick *gesture,
@@ -1407,18 +1613,8 @@ released_gesture_helper (HexWidget *self,
 		double			y,
 		HexWidgetViewType type)
 {
-	GtkWidget *widget;
-	guint button;
-
-	if (type == VIEW_HEX)
-		widget = self->xdisp;
-	else
-		widget = self->adisp;
-
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	g_return_if_fail (GTK_IS_WIDGET(widget));
-	
-	button = gtk_gesture_single_get_current_button (GTK_GESTURE_SINGLE(gesture));
+	guint button = gtk_gesture_single_get_current_button (
+			GTK_GESTURE_SINGLE(gesture));
 
 	/* Single-click */
 	if (button == GDK_BUTTON_PRIMARY && n_press == 1)
@@ -1429,9 +1625,18 @@ released_gesture_helper (HexWidget *self,
 			self->scroll_timeout = 0;
 			self->scroll_dir = 0;
 		}
+
+		update_primary_selection (self);
 		self->selecting = FALSE;
 		self->button = 0;
 	}
+	/* Single-click */
+	else if (button == GDK_BUTTON_MIDDLE && n_press == 1)
+	{
+		GdkClipboard *primary = gtk_widget_get_primary_clipboard (GTK_WIDGET(self));
+		g_debug ("%s: middle-click paste - TEST", __func__);
+		paste_helper (self, primary);
+	}
 }
 
 /* nb: this gesture is only associated with the right-click, so there is
@@ -1492,9 +1697,6 @@ drag_update_helper (HexWidget *self,
 	else
 		widget = self->adisp;
 
-	g_return_if_fail (HEX_IS_WIDGET(self));
-	g_return_if_fail (GTK_IS_WIDGET(widget));
-
 	gtk_widget_get_allocation (widget, &allocation);
 	gtk_gesture_drag_get_start_point (gesture, &start_x, &start_y);
 
@@ -1641,14 +1843,20 @@ key_press_cb (GtkEventControllerKey *controller,
 			break;
 
 		case GDK_KEY_Page_Up:
-			hex_widget_set_cursor (self, MAX (0, self->cursor_pos - self->vis_lines*self->cpl));
-			ret = GDK_EVENT_STOP;
+			if (! (state & GDK_CONTROL_MASK))
+			{
+				hex_widget_set_cursor (self, MAX (0, self->cursor_pos - self->vis_lines*self->cpl));
+				ret = GDK_EVENT_STOP;
+			}
 			break;
 
 		case GDK_KEY_Page_Down:
-			hex_widget_set_cursor(self, MIN (payload_size,
-						self->cursor_pos + self->vis_lines*self->cpl));
-			ret = GDK_EVENT_STOP;
+			if (! (state & GDK_CONTROL_MASK))
+			{
+				hex_widget_set_cursor(self, MIN (payload_size,
+							self->cursor_pos + self->vis_lines*self->cpl));
+				ret = GDK_EVENT_STOP;
+			}
 			break;
 
 		case GDK_KEY_Home:
@@ -1837,22 +2045,6 @@ key_release_cb (GtkEventControllerKey *controller,
 	return ret;
 }
 
-static void
-show_offsets_widget(HexWidget *self)
-{
-	g_return_if_fail (GTK_IS_WIDGET (self->offsets));
-
-	gtk_widget_show (self->offsets);
-}
-
-static void
-hide_offsets_widget(HexWidget *self)
-{
-	g_return_if_fail (gtk_widget_get_realized (self->offsets));
-
-	gtk_widget_hide (self->offsets);
-}
-
 /* FIXME - Reorganize/clean up. Mostly flown in from old code.
  */
 /*
@@ -1907,12 +2099,7 @@ hex_widget_real_data_changed (HexWidget *self, gpointer data)
 	else
 		end_line = MIN(end_line, self->vis_lines);
 
-    invalidate_hex_lines (self, start_line, end_line);
-    invalidate_ascii_lines (self, start_line, end_line);
-    if (self->show_offsets)
-    {
-        invalidate_offsets (self, start_line, end_line);
-    }
+	gtk_widget_queue_draw (GTK_WIDGET(self));
 }
 
 static void
@@ -1933,13 +2120,7 @@ bytes_changed (HexWidget *self, gint64 start, gint64 end)
 	if (end_line < 0 || start_line > self->vis_lines)
 		return;
 
-    invalidate_hex_lines (self, start_line, end_line);
-    invalidate_ascii_lines (self, start_line, end_line);
-
-    if (self->show_offsets)
-    {
-        invalidate_offsets (self, start_line, end_line);
-    }
+	gtk_widget_queue_draw (GTK_WIDGET(self));
 }
 
 static void
@@ -2022,41 +2203,37 @@ hex_widget_delete_highlight (HexWidget *self, HexWidgetAutoHighlight *ahl,
 	bytes_changed(self, start, end);
 }
 
-/* stolen from hex_document_compare_data - but uses
- * hex_widget_* stuff rather than hex_document_* directly
- * and simply returns a gboolean.
- */
-static gboolean
-hex_widget_compare_data (HexWidget *self, guchar *cmp, gint64 pos, int len)
-{
-	for (int i = 0; i < len; i++)
-	{
-		guchar c = hex_widget_get_byte (self, pos + i);
-
-		if (c != *(cmp + i))
-			return FALSE;
-	}
-	return TRUE;
-}
-
 static gboolean
 hex_widget_find_limited (HexWidget *self, char *find, int findlen,
-		gint64 lower, gint64 upper, gint64 *found)
+		HexSearchFlags flags, gint64 lower, gint64 upper,
+		gint64 *found, size_t *found_len)
 {
+	gboolean retval = FALSE;
 	gint64 pos = lower;
+	HexDocumentFindData *find_data = hex_document_find_data_new ();
+
+	find_data->what = find;
+	find_data->len = findlen;
+	find_data->flags = flags;
 
 	while (pos < upper)
 	{
-		if (hex_widget_compare_data (self, (guchar *)find, pos, findlen))
+		if (hex_document_compare_data_full (self->document, find_data, pos)
+				== 0)
 		{
 			*found = pos;
-			return TRUE;
+			*found_len = find_data->found_len;
+
+			retval = TRUE;
+			goto out;
 		}
 		pos++;
 	}
-	return FALSE;
-}
 
+out:
+	g_free (find_data);
+	return retval;
+}
 
 /* Helper for the autohighlight functions to set a reasonable default
  * view_min () and view_max () value.
@@ -2075,6 +2252,7 @@ hex_widget_update_auto_highlight (HexWidget *self, HexWidgetAutoHighlight *ahl,
 	gint64 del_min, del_max;
 	gint64 add_min, add_max;
 	gint64 foundpos = -1;
+	size_t found_len;
 	gint64 prev_min = ahl->view_min;
 	gint64 prev_max = ahl->view_max;
 	HexWidget_Highlight *cur;
@@ -2116,15 +2294,23 @@ hex_widget_update_auto_highlight (HexWidget *self, HexWidgetAutoHighlight *ahl,
 			hex_widget_delete_highlight (self, ahl, cur);
 			cur = next;
 		}
-		else cur = cur->next;
+		else
+			cur = cur->next;
 	}
 
 	while (add &&
-		   hex_widget_find_limited (self, ahl->search_string, ahl->search_len,
-			   /* lower */ MAX(add_min, foundpos+1), add_max, &foundpos))
+		   hex_widget_find_limited (self,
+			   ahl->search_string,
+			   ahl->search_len,
+			   ahl->search_flags,
+			   MAX(add_min, foundpos+1),	/* lower */
+			   add_max,						/* upper */
+			   &foundpos,
+			   &found_len))
 	{
 		hex_widget_insert_highlight (self, ahl,
-				foundpos, foundpos+(ahl->search_len)-1);
+				foundpos,
+				foundpos + found_len - 1);
 	}
 }
 
@@ -2203,119 +2389,28 @@ static void
 hex_widget_real_cut_to_clipboard(HexWidget *self,
 		gpointer user_data)
 {
-	(void)user_data;
-
 	if (self->selection.start != -1 && self->selection.end != -1) {
 		hex_widget_real_copy_to_clipboard(self);
 		hex_widget_delete_selection(self);
 	}
 }
 
-static void
-plaintext_paste_received_cb (GObject *source_object,
-		GAsyncResult *result,
-		gpointer user_data)
-{
-	HexWidget *self = HEX_WIDGET(user_data);
-	GdkClipboard *clipboard;
-	char *text;
-	GError *error = NULL;
-
-	g_debug ("%s: We DON'T have our special HexPasteData. Falling back "
-			"to plaintext paste.",
-			__func__);
-
-	clipboard = GDK_CLIPBOARD (source_object);
-
-	/* Get the resulting text of the read operation */
-	text = gdk_clipboard_read_text_finish (clipboard, result, &error);
-
-	if (text) {
-		hex_document_set_data (self->document,
-				self->cursor_pos,
-				strlen(text),
-				0,	/* rep_len (0 to insert w/o replacing; what we want) */
-				text,
-				TRUE);
-
-		hex_widget_set_cursor (self, self->cursor_pos + strlen(text));
-		
-		g_free(text);
-	}
-	else {
-		g_critical ("Error pasting text: %s", 
-				error->message);
-		g_error_free (error);
-	}
-}
-
 static void
 hex_widget_real_paste_from_clipboard (HexWidget *self,
 		gpointer user_data)
 {
 	GtkWidget *widget = GTK_WIDGET(self);
 	GdkClipboard *clipboard;
-	GdkContentProvider *content;
-	GValue value = G_VALUE_INIT;
-	HexPasteData *paste;
-	gboolean have_hex_paste_data = FALSE;
-
-	(void)user_data;	/* unused */
 
 	clipboard = gtk_widget_get_clipboard (widget);
-	content = gdk_clipboard_get_content (clipboard);
-	g_value_init (&value, HEX_TYPE_PASTE_DATA);
 
-	/* If the clipboard contains our special HexPasteData, we'll use it.
-	 * If not, just fall back to plaintext.
-	 *
-	 * Note the double test here; it seems the test is semi-superfluous for
-	 * *this* purpose because _get_content will itself return NULL if the
-	 * clipboard data we're getting is not owned by the process; that will
-	 * pretty much *always* be the case when we're falling back to plaintext,
-	 * ie, when pasting from external apps. Oh well.
-	 */
-	have_hex_paste_data =
-		GDK_IS_CONTENT_PROVIDER (content) &&
-		gdk_content_provider_get_value (content,
-				&value,
-				NULL);	/* GError - NULL to ignore */
-
-	if (have_hex_paste_data)
-	{
-		char *doc_data;
-		int elems;
-
-		g_debug("%s: We HAVE our special HexPasteData.",
-				__func__);
-
-		paste = HEX_PASTE_DATA(g_value_get_object (&value));
-		doc_data = hex_paste_data_get_doc_data (paste);
-		elems = hex_paste_data_get_elems (paste);
-
-		hex_document_set_data (self->document,
-				self->cursor_pos,
-				elems,
-				0,	/* rep_len (0 to insert w/o replacing; what we want) */
-				doc_data,
-				TRUE);
-
-		hex_widget_set_cursor (self, self->cursor_pos + elems);
-	}
-	else {
-		gdk_clipboard_read_text_async (clipboard,
-				NULL,	/* GCancellable *cancellable */
-				plaintext_paste_received_cb,
-				self);
-	}
+	paste_helper (self, clipboard);
 }
 
 static void
 hex_widget_real_draw_complete (HexWidget *self,
 		gpointer user_data)
 {
-	(void)user_data;
-
 	recalc_scrolling (self);
 }
 
@@ -2361,7 +2456,8 @@ hex_widget_snapshot (GtkWidget *widget, GtkSnapshot *snapshot)
 			child != NULL;
 			child = gtk_widget_get_next_sibling (child))
 	{
-		gtk_widget_snapshot_child (widget, child, snapshot);
+		if (gtk_widget_get_visible (child))
+			gtk_widget_snapshot_child (widget, child, snapshot);
 	}
 
 	g_signal_emit_by_name(G_OBJECT(self), "draw-complete");
@@ -2418,7 +2514,6 @@ document_changed_cb (HexDocument* doc, gpointer change_data,
         gboolean push_undo, gpointer data)
 {
 	HexWidget *self = HEX_WIDGET(data);
-	g_return_if_fail (HEX_IS_WIDGET (self));
 
     hex_widget_real_data_changed (self, change_data);
 
@@ -2428,6 +2523,32 @@ document_changed_cb (HexDocument* doc, gpointer change_data,
 			"gtkhex.redo", hex_document_can_redo (doc));
 }
 
+static void
+hex_widget_constructed (GObject *object)
+{
+	HexWidget *self = HEX_WIDGET(object);
+
+	/* Setup document signals */
+
+	g_signal_connect (self->document, "file-read-started",
+			G_CALLBACK (file_read_started_cb), self);
+
+	g_signal_connect (self->document, "file-loaded",
+			G_CALLBACK (file_loaded_cb), self);
+
+	g_signal_connect (self->document, "document-changed",
+			G_CALLBACK (document_changed_cb), self);
+
+	g_signal_connect (self->document, "undo",
+			G_CALLBACK (doc_undo_redo_cb), self);
+
+	g_signal_connect (self->document, "redo",
+			G_CALLBACK (doc_undo_redo_cb), self);
+
+	/* Chain up */
+	G_OBJECT_CLASS(hex_widget_parent_class)->constructed (object);
+}
+
 static void
 hex_widget_dispose (GObject *object)
 {
@@ -2443,10 +2564,6 @@ hex_widget_dispose (GObject *object)
 	g_clear_pointer (&self->scrollbar, gtk_widget_unparent);
 	g_clear_pointer (&self->busy_spinner, gtk_widget_unparent);
 
-	/* FIXME - This results in assertion errors upon exit. I have been told
-	 * by ebassi on IRC (16-Dec-2021) that this may be a bug in gtk. See:
-	 * https://gitlab.gnome.org/GNOME/gtk/-/issues/4548
-	 */
 	g_clear_pointer (&self->context_menu, gtk_widget_unparent);
 
 	/* Clear pango layouts
@@ -2483,11 +2600,23 @@ hex_widget_class_init (HexWidgetClass *klass)
 
 	/* vfuncs */
 
+	object_class->constructed = hex_widget_constructed;
 	object_class->dispose = hex_widget_dispose;
 	object_class->finalize = hex_widget_finalize;
+	object_class->set_property = hex_widget_set_property;
+	object_class->get_property = hex_widget_get_property;
+
 	widget_class->snapshot = hex_widget_snapshot;
 
-	/* Layout manager: box-style layout. */
+	/* PROPERTIES */
+
+	properties[DOCUMENT] = g_param_spec_object ("document", NULL, NULL,
+			HEX_TYPE_DOCUMENT,
+			G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+
+	g_object_class_install_properties (object_class, N_PROPERTIES, properties);
+
+	/* Layout manager */
 
 	gtk_widget_class_set_layout_manager_type (widget_class,
 			HEX_TYPE_WIDGET_LAYOUT);
@@ -2727,7 +2856,7 @@ hex_widget_init (HexWidget *self)
 
 	self->provider = gtk_css_provider_new ();
 	gtk_css_provider_load_from_resource (GTK_CSS_PROVIDER (self->provider),
-		RESOURCE_BASE_PATH "/css/ghex.css");
+		RESOURCE_BASE_PATH "/css/gtkhex.css");
 
 	gtk_style_context_add_provider (context, GTK_STYLE_PROVIDER (self->provider),
 			GTK_STYLE_PROVIDER_PRIORITY_THEME);
@@ -2826,7 +2955,6 @@ hex_widget_init (HexWidget *self)
 	gtk_widget_set_parent (self->context_menu, widget);
 
 	g_object_unref (builder);
-	g_object_unref (menu);		/* ref'd by gtk_popover_menu_new_from_model */
 
 	/* Initialize Adjustment */
 	self->adj = gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
@@ -2854,8 +2982,6 @@ hex_widget_init (HexWidget *self)
 	gtk_widget_set_visible (self->busy_spinner, FALSE);
 	gtk_widget_set_parent (self->busy_spinner, widget);
 
-	/* SETUP SIGNALS */
-
 	/* GESTURES */
 
 	/* Whole HexWidget widget (for right-click context menu) */
@@ -2973,6 +3099,11 @@ hex_widget_init (HexWidget *self)
 			"gtkhex.undo", FALSE);
 	gtk_widget_action_set_enabled (GTK_WIDGET(self),
 			"gtkhex.redo", FALSE);
+
+	/* PRIMARY selection content */
+
+	self->selection_content = hex_content_provider_new ();
+	HEX_CONTENT_PROVIDER (self->selection_content)->owner = self;
 }
 
 /*-------- public API starts here --------*/
@@ -2989,31 +3120,9 @@ hex_widget_init (HexWidget *self)
 GtkWidget *
 hex_widget_new (HexDocument *owner)
 {
-	HexWidget *self;
-
-	self = HEX_WIDGET (g_object_new (HEX_TYPE_WIDGET, NULL));
-	g_return_val_if_fail (self != NULL, NULL);
-
-	self->document = owner;
-
-	/* Setup document signals */
-
-    g_signal_connect (G_OBJECT (self->document), "file-read-started",
-            G_CALLBACK (file_read_started_cb), self);
-
-    g_signal_connect (G_OBJECT (self->document), "file-loaded",
-            G_CALLBACK (file_loaded_cb), self);
-
-    g_signal_connect (G_OBJECT (self->document), "document-changed",
-            G_CALLBACK (document_changed_cb), self);
-
-    g_signal_connect (G_OBJECT (self->document), "undo",
-            G_CALLBACK (doc_undo_redo_cb), self);
-	
-    g_signal_connect (G_OBJECT (self->document), "redo",
-            G_CALLBACK (doc_undo_redo_cb), self);
-	
-	return GTK_WIDGET(self);
+	return g_object_new (HEX_TYPE_WIDGET,
+			"document", owner,
+			NULL);
 }
 
 /**
@@ -3393,11 +3502,48 @@ hex_widget_show_offsets (HexWidget *self, gboolean show)
 {
 	g_return_if_fail (HEX_IS_WIDGET (self));
 
+	gtk_widget_set_visible (self->offsets, show);
 	self->show_offsets = show;
-	if (show)
-		show_offsets_widget(self);
-	else
-		hide_offsets_widget(self);
+}
+
+/**
+ * hex_widget_show_hex_column:
+ * @show: %TRUE if the hex column should be shown, %FALSE if it should
+ *   be hidden
+ *
+ * Set whether the hex column of the widget should be shown.
+ *
+ * Since: 4.2
+ */
+void
+hex_widget_show_hex_column (HexWidget *self, gboolean show)
+{
+	g_return_if_fail (HEX_IS_WIDGET (self));
+
+	if (!show && gtk_widget_get_visible (self->adisp))
+		self->active_view = VIEW_ASCII;
+
+	gtk_widget_set_visible (self->xdisp, show);
+}
+
+/**
+ * hex_widget_show_ascii_column:
+ * @show: %TRUE if the ASCII column should be shown, %FALSE if it should
+ *   be hidden
+ *
+ * Set whether the ASCII column of the widget should be shown.
+ *
+ * Since: 4.2
+ */
+void
+hex_widget_show_ascii_column (HexWidget *self, gboolean show)
+{
+	g_return_if_fail (HEX_IS_WIDGET (self));
+
+	if (!show && gtk_widget_get_visible (self->xdisp))
+		self->active_view = VIEW_HEX;
+
+	gtk_widget_set_visible (self->adisp, show);
 }
 
 /**
@@ -3422,23 +3568,30 @@ hex_widget_set_insert_mode (HexWidget *self, gboolean insert)
 }
 
 /**
- * hex_widget_insert_autohighlight:
+ * hex_widget_insert_autohighlight_full:
  * @search: (array length=len) (transfer full): search string to auto-highlight
  * @len: length of the @search string
- *   
- * Insert an auto-highlight of a given search string.
+ * @flags: #HexSearchFlags to specify match type
+ *
+ * Full version of [method@Hex.Widget.insert_autohighlight] which allows
+ * for specifying string match types for auto-highlights over and above
+ * exact byte-for-byte string matches.
  *
  * Returns: a newly created [struct@Hex.WidgetAutoHighlight] structure
+ *
+ * Since: 4.2
  */
 HexWidgetAutoHighlight *
-hex_widget_insert_autohighlight (HexWidget *self,
+hex_widget_insert_autohighlight_full (HexWidget *self,
 		const char *search,
-		int len)
+		int len,
+		HexSearchFlags flags)
 {
 	HexWidgetAutoHighlight *new = g_new0 (HexWidgetAutoHighlight, 1);
 
 	new->search_string = g_memdup2 (search, len);
 	new->search_len = len;
+	new->search_flags = flags;
 
 	new->highlights = NULL;
 
@@ -3454,6 +3607,24 @@ hex_widget_insert_autohighlight (HexWidget *self,
 	return new;
 }
 
+/**
+ * hex_widget_insert_autohighlight:
+ * @search: (array length=len) (transfer full): search string to auto-highlight
+ * @len: length of the @search string
+ *
+ * Insert an auto-highlight of a given search string.
+ *
+ * Returns: a newly created [struct@Hex.WidgetAutoHighlight] structure
+ */
+HexWidgetAutoHighlight *
+hex_widget_insert_autohighlight (HexWidget *self,
+		const char *search,
+		int len)
+{
+	return hex_widget_insert_autohighlight_full (
+			self, search, len, HEX_SEARCH_NONE);
+}
+
 /* FIXME - use _free func. */
 /**
  * hex_widget_delete_autohighlight:
