@@ -52,6 +52,9 @@ static void undo_stack_ascend           (HexDocument *doc);
 static void undo_stack_free             (HexDocument *doc);
 
 #define DEFAULT_UNDO_DEPTH 1024
+#define REGEX_SEARCH_LEN 1024	/* FIXME/TODO: This is kind of lazy. Stopgap? */
+
+/* SIGNALS */
 
 enum {
 	DOCUMENT_CHANGED,
@@ -67,23 +70,56 @@ enum {
 
 static guint hex_signals[LAST_SIGNAL];
 
+/* PROPERTIES */
+
+enum
+{
+	GFILE = 1,
+	BUFFER,
+	N_PROPERTIES
+};
+
+static GParamSpec *properties[N_PROPERTIES];
 
-/* TODO - come up with _new functions for the GTypes below and actually use
- * them. */
 
 /* HexDocumentFindData GType Definitions */
 
-/* FIXME - unused and could be unreliable. */
-static HexDocumentFindData *
+/**
+ * hex_document_find_data_new:
+ *
+ * Create a new empty [struct@Hex.DocumentFindData] structure.
+ *
+ * Returns: a new #HexDocumentFindData structure. Can be freed with
+ *   `g_free ()`.
+ *
+ * Since: 4.2
+ */
+HexDocumentFindData *
+hex_document_find_data_new (void)
+{
+	return g_new0 (HexDocumentFindData, 1);
+}
+
+/**
+ * hex_document_find_data_copy:
+ *
+ * Copy a [struct@Hex.DocumentFindData] structure. This function is likely
+ * only useful for language bindings.
+ *
+ * Returns: a newly allocated #HexDocumentFindData structure. Can be freed with
+ *   `g_free ()`.
+ *
+ * Since: 4.2
+ */
+HexDocumentFindData *
 hex_document_find_data_copy (HexDocumentFindData *data)
 {
-	return data;
+	return g_memdup2 (data, sizeof *data);
 }
 
 G_DEFINE_BOXED_TYPE (HexDocumentFindData, hex_document_find_data,
 		hex_document_find_data_copy, g_free)
 
-
 /* HexChangeData GType Definitions */
 
 /* FIXME - unused and could be unreliable */
@@ -122,6 +158,58 @@ struct _HexDocument
 
 G_DEFINE_TYPE (HexDocument, hex_document, G_TYPE_OBJECT)
 
+
+/* PROPERTIES - GETTERS AND SETTERS */
+
+static void
+hex_document_set_property (GObject *object,
+		guint property_id,
+		const GValue *value,
+		GParamSpec *pspec)
+{
+	HexDocument *doc = HEX_DOCUMENT(object);
+
+	switch (property_id)
+	{
+		case GFILE:
+			hex_document_set_file (doc, g_value_get_object (value));
+			break;
+
+		case BUFFER:
+			hex_document_set_buffer (doc, g_value_get_object (value));
+			break;
+
+		default:
+			/* We don't have any other property... */
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+			break;
+	}
+}
+
+static void
+hex_document_get_property (GObject *object,
+		guint property_id,
+		GValue *value,
+		GParamSpec *pspec)
+{
+	HexDocument *doc = HEX_DOCUMENT(object);
+
+	switch (property_id)
+	{
+		case GFILE:
+			/* --- */
+			break;
+
+		case BUFFER:
+			break;
+
+		default:
+			/* We don't have any other property... */
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+			break;
+	}
+}
+
 /* ---- */
 
 static void
@@ -253,9 +341,27 @@ static void
 hex_document_class_init (HexDocumentClass *klass)
 {
 	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
+	GParamFlags default_flags = G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS |
+		G_PARAM_EXPLICIT_NOTIFY;
 	
 	gobject_class->finalize = hex_document_finalize;
-	gobject_class->dispose= hex_document_dispose;
+	gobject_class->dispose = hex_document_dispose;
+	gobject_class->set_property = hex_document_set_property;
+	gobject_class->get_property = hex_document_get_property;
+
+	/* PROPERTIES */
+
+	properties[GFILE] = g_param_spec_object ("file", NULL, NULL, 
+			G_TYPE_FILE,
+			default_flags);
+
+	properties[BUFFER] = g_param_spec_object ("buffer", NULL, NULL, 
+			HEX_TYPE_BUFFER,
+			default_flags);
+
+	g_object_class_install_properties (gobject_class, N_PROPERTIES, properties);
+
+	/* SIGNALS */
 	
 	hex_signals[DOCUMENT_CHANGED] =
 		g_signal_new_class_handler ("document-changed",
@@ -333,11 +439,17 @@ hex_document_class_init (HexDocumentClass *klass)
 static void
 hex_document_init (HexDocument *doc)
 {
-#ifdef BACKEND_MMAP
-	doc->buffer = hex_buffer_util_new ("mmap", NULL);
-#else
-	doc->buffer = hex_buffer_util_new (NULL, NULL);
-#endif
+	HexBuffer *try_buf = NULL;
+
+	/* FIXME - make this a setting; preferred default backend, or
+	 * what-have-you.
+	 */
+	try_buf = hex_buffer_util_new ("mmap", NULL);
+	if (! try_buf)
+		try_buf = hex_buffer_util_new (NULL, NULL);
+
+	g_assert (try_buf != NULL);
+	doc->buffer = try_buf;
 
 	doc->undo_max = DEFAULT_UNDO_DEPTH;
 }
@@ -380,7 +492,9 @@ hex_document_set_file (HexDocument *doc, GFile *file)
 	}
 
 	doc->file = g_object_ref (file);
+
 	g_signal_emit (G_OBJECT(doc), hex_signals[FILE_NAME_CHANGED], 0);
+	g_object_notify_by_pspec (G_OBJECT(doc), properties[BUFFER]);
 
 	return TRUE;
 }
@@ -601,13 +715,10 @@ document_ready_cb (GObject *source_object,
 		GAsyncResult *res,
 		gpointer user_data)
 {
-	static HexChangeData change_data;
-
 	gboolean success;
 	GError *local_error = NULL;
 	HexBuffer *buf = HEX_BUFFER(source_object);
 	GTask *task = G_TASK (user_data);
-	gint64 payload;
 	HexDocument *doc;
 
 	doc = HEX_DOCUMENT(g_task_get_task_data (task));
@@ -628,13 +739,6 @@ document_ready_cb (GObject *source_object,
 
 	undo_stack_free(doc);
 
-	payload = hex_buffer_get_payload_size (hex_document_get_buffer (doc));
-
-	change_data.start = 0;
-	change_data.end = payload - 1;
-
-	doc->changed = FALSE;
-	hex_document_changed (doc, &change_data, FALSE);
 	g_signal_emit (G_OBJECT(doc), hex_signals[FILE_LOADED], 0);
 	g_task_return_boolean (task, TRUE);
 
@@ -658,7 +762,7 @@ cleanup:
  */
 void
 hex_document_read_async (HexDocument *doc,
-		GCancellable *cancellable,	/* FIXME: presently ignored */
+		GCancellable *cancellable,
 		GAsyncReadyCallback callback,
 		gpointer user_data)
 {
@@ -763,6 +867,7 @@ write_ready_cb (GObject *source_object,
 {
 	HexBuffer *buf = HEX_BUFFER(source_object);
 	GTask *doc_task = G_TASK(user_data);
+	HexDocument *doc = g_task_get_source_object (doc_task);
 	gboolean success;
 	GError *local_error = NULL;
 
@@ -771,6 +876,8 @@ write_ready_cb (GObject *source_object,
 
 	if (success)
 	{
+		doc->changed = FALSE;
+		g_signal_emit (doc, hex_signals[FILE_SAVED], 0);
 		g_task_return_boolean (doc_task, TRUE);
 	}
 	else
@@ -802,7 +909,7 @@ write_ready_cb (GObject *source_object,
  */
 void
 hex_document_write_async (HexDocument *doc,
-		GCancellable *cancellable,	/* FIXME: presently ignored */
+		GCancellable *cancellable,
 		GAsyncReadyCallback callback,
 		gpointer user_data)
 {
@@ -846,13 +953,13 @@ out:
 void
 hex_document_write_to_file_async (HexDocument *doc,
 		GFile *file,
-		GCancellable *cancellable,	/* FIXME: presently ignored */
+		GCancellable *cancellable,
 		GAsyncReadyCallback callback,
 		gpointer user_data)
 {
 	GTask *doc_task;
 
-	g_return_if_fail (G_IS_FILE (doc->file));
+	g_return_if_fail (G_IS_FILE (file));
 
 	doc_task = g_task_new (doc, cancellable, callback, user_data);
 
@@ -1095,6 +1202,116 @@ hex_document_export_html (HexDocument *doc,
 	return TRUE;
 }
 
+/**
+ * hex_document_compare_data_full:
+ * @find_data: a #HexDocumentFindData structure
+ * @pos: offset position of the #HexDocument data to compare with the
+ *   string contained in the `find_data` structure
+ *
+ * Full version of [method@Hex.Document.compare_data] to allow data
+ * comparisons broader than byte-for-byte matches only. However, it is
+ * less convenient than the above since it requires the caller to allocate
+ * and free a #HexDocumentFindData structure.
+ *
+ * Returns: 0 if the comparison is an exact match; otherwise, a non-zero
+ *   value is returned.
+ *
+ * Since: 4.2
+ */
+int
+hex_document_compare_data_full (HexDocument *doc,
+		HexDocumentFindData *find_data,
+		gint64 pos)
+{
+	char *cp = 0;
+	GError *local_error = NULL;
+	int retval = 1;		/* match will make it zero, so set it non-zero now. */
+
+	g_return_val_if_fail (find_data, 0);
+	g_return_val_if_fail (find_data->what, 0);
+
+	if (find_data->flags & HEX_SEARCH_REGEX)
+	{
+		GRegex *regex;
+		GMatchInfo *match_info;
+		char *regex_search_str;
+		GRegexCompileFlags regex_compile_flags;
+
+		/* GRegex doesn't let you specify the length of the search string, so
+		 * it needs to be NULL-terminated.
+		 */
+		regex_search_str = g_malloc (find_data->len+1);
+		memcpy (regex_search_str, find_data->what, find_data->len);
+		regex_search_str[find_data->len] = 0;
+
+		/* match string doesn't have to be UTF-8 */
+		regex_compile_flags = G_REGEX_RAW;
+
+		if (find_data->flags & HEX_SEARCH_IGNORE_CASE)
+			regex_compile_flags |= G_REGEX_CASELESS;
+
+		regex = g_regex_new (regex_search_str,
+				regex_compile_flags,
+				G_REGEX_MATCH_ANCHORED,
+				&local_error);
+
+		g_free (regex_search_str);
+
+		/* sanity check */
+		if (!regex || local_error)
+		{
+			g_debug ("%s: error: %s", __func__, local_error->message);
+			goto out;
+		}
+
+		cp = hex_buffer_get_data (doc->buffer, pos, REGEX_SEARCH_LEN);
+
+		if (g_regex_match_full (regex, cp,
+					REGEX_SEARCH_LEN,	/* length of string being searched */
+					0,					/* start pos */
+					0,					/* addl match_options */
+					&match_info,
+					&local_error))
+		{
+			char *word = g_match_info_fetch (match_info, 0);
+
+			find_data->found_len = strlen (word);
+			g_free (word);
+			retval = 0;
+		}
+		else
+		{
+			if (local_error)
+			{
+				g_debug ("%s: error: %s",
+						__func__,
+						local_error ? local_error->message : NULL);
+			}
+			retval = 1;
+		}
+	}
+	else	/* non regex */
+	{
+		cp = hex_buffer_get_data (doc->buffer, pos, find_data->len);
+
+		if (find_data->flags & HEX_SEARCH_IGNORE_CASE)
+		{
+			retval = g_ascii_strncasecmp (cp, find_data->what, find_data->len);
+		}
+		else
+		{
+			retval = memcmp (cp, find_data->what, find_data->len);
+		}
+
+		if (retval == 0)
+			find_data->found_len = find_data->len;
+	}
+out:
+	g_clear_error (&local_error);
+	g_free (cp);
+	return retval;
+}
+
 /**
  * hex_document_compare_data:
  * @doc: a [class@Hex.Document] object
@@ -1110,19 +1327,58 @@ int
 hex_document_compare_data (HexDocument *doc,
 		const char *what, gint64 pos, size_t len)
 {
-	char c;
+	int retval;
+	HexDocumentFindData *find_data = hex_document_find_data_new ();
 
-	g_return_val_if_fail (what, 0);
+	find_data->what = what;
+	find_data->len = len;
+	find_data->flags = HEX_SEARCH_NONE;
 
-	for (size_t i = 0; i < len; i++, what++)
+	retval = hex_document_compare_data_full (doc, find_data, pos);
+	g_free (find_data);
+
+	return retval;
+}
+
+/**
+ * hex_document_find_forward_full:
+ * @find_data: a #HexDocumentFindData structure
+ *
+ * Full version of [method@Hex.Document.find_forward] which allows for
+ * more flexibility than the above, which is only for a byte-by-byte exact
+ * match. However, it is less convenient to call since the caller must
+ * create and and free a #HexDocumentFindData structure manually.
+ *
+ * This method will block. For a non-blocking version, use
+ * [method@Hex.Document.find_forward_async].
+ *
+ * Returns: %TRUE if the search string contained in `find_data` was found by
+ *   the requested operation; %FALSE otherwise.
+ *
+ * Since: 4.2
+ */
+gboolean
+hex_document_find_forward_full (HexDocument *doc,
+		HexDocumentFindData *find_data)
+{
+	gint64 pos;
+	gint64 payload = hex_buffer_get_payload_size (
+			hex_document_get_buffer (doc));
+
+	g_return_val_if_fail (find_data != NULL, FALSE);
+
+	pos = find_data->start;
+	while (pos < payload)
 	{
-		c = hex_buffer_get_byte (doc->buffer, pos + i);
-
-		if (c != *what)
-			return (c - *what);
+		if (hex_document_compare_data_full (doc, find_data, pos) == 0)
+		{
+			find_data->offset = pos;
+			return TRUE;
+		}
+		pos++;
 	}
-	
-	return 0;
+
+	return FALSE;
 }
 
 /**
@@ -1148,22 +1404,20 @@ gboolean
 hex_document_find_forward (HexDocument *doc, gint64 start, const char *what,
 						  size_t len, gint64 *offset)
 {
-	gint64 pos;
-	gint64 payload = hex_buffer_get_payload_size (
-			hex_document_get_buffer (doc));
+	gboolean retval;
+	HexDocumentFindData *find_data = hex_document_find_data_new ();
 
-	pos = start;
-	while (pos < payload)
-	{
-		if (hex_document_compare_data (doc, what, pos, len) == 0)
-		{
-			*offset = pos;
-			return TRUE;
-		}
-		pos++;
-	}
+	find_data->start = start;
+	find_data->what = what;
+	find_data->len = len;
+	find_data->flags = HEX_SEARCH_NONE;
 
-	return FALSE;
+	retval = hex_document_find_forward_full (doc, find_data);
+	*offset = find_data->offset;
+
+	g_free (find_data);
+
+	return retval;
 }
 
 /**
@@ -1185,6 +1439,26 @@ hex_document_find_finish (HexDocument *doc,
 	return g_task_propagate_pointer (G_TASK(result), NULL);
 }
 
+#define FIND_FULL_THREAD_TEMPLATE(FUNC_NAME, FUNC_TO_CALL) \
+static void \
+FUNC_NAME (GTask *task, \
+		gpointer source_object, \
+		gpointer task_data, \
+		GCancellable *cancellable) \
+{ \
+	HexDocument *doc = HEX_DOCUMENT (source_object); \
+	HexDocumentFindData *find_data = task_data; \
+ \
+	g_return_if_fail (find_data); \
+ \
+	find_data->found = FUNC_TO_CALL (doc, find_data); \
+ \
+	g_task_return_pointer (task, find_data, g_free); \
+}
+
+FIND_FULL_THREAD_TEMPLATE(hex_document_find_forward_full_thread,
+		hex_document_find_forward_full)
+
 static void
 hex_document_find_forward_thread (GTask *task,
 		gpointer source_object,
@@ -1201,7 +1475,36 @@ hex_document_find_forward_thread (GTask *task,
 	g_task_return_pointer (task, find_data, g_free);
 }
 
-/* CROSSREF: hex-document.h - HexDocumentFindData */
+#define FIND_FULL_ASYNC_TEMPLATE(FUNC_NAME, FUNC_TO_CALL) \
+void \
+FUNC_NAME (HexDocument *doc, \
+		HexDocumentFindData *find_data, \
+		GCancellable *cancellable, \
+		GAsyncReadyCallback callback, \
+		gpointer user_data) \
+{ \
+	GTask *task; \
+ \
+	task = g_task_new (doc, cancellable, callback, user_data); \
+	g_task_set_return_on_cancel (task, TRUE); \
+	g_task_set_task_data (task, find_data, g_free); \
+	g_task_run_in_thread (task, FUNC_TO_CALL); \
+}
+
+/**
+ * hex_document_find_forward_full_async:
+ * @find_data: a #HexDocumentFindData structure
+ * @cancellable: (nullable): a #GCancellable
+ * @callback: (scope async): function to be called when the operation is
+ *   complete
+ *
+ * Non-blocking version of [method@Hex.Document.find_forward_full].
+ *
+ * Since: 4.2
+ */
+FIND_FULL_ASYNC_TEMPLATE(hex_document_find_forward_full_async, 
+		hex_document_find_forward_full_thread)
+
 /**
  * hex_document_find_forward_async:
  * @doc: a [class@Hex.Document] object
@@ -1221,31 +1524,74 @@ hex_document_find_forward_thread (GTask *task,
  * function that should generally be used by a GUI client to find a string
  * forwards in a #HexDocument.
  */
-void
-hex_document_find_forward_async (HexDocument *doc,
-		gint64 start,
-		const char *what,
-		size_t len,
-		gint64 *offset,
-		const char *found_msg,
-		const char *not_found_msg,
-		GCancellable *cancellable,
-		GAsyncReadyCallback callback,
-		gpointer user_data)
+
+#define FIND_ASYNC_TEMPLATE(FUNC_NAME, FUNC_TO_CALL) \
+void \
+FUNC_NAME (HexDocument *doc, \
+		gint64 start, \
+		const char *what, \
+		size_t len, \
+		gint64 *offset, \
+		const char *found_msg, \
+		const char *not_found_msg, \
+		GCancellable *cancellable, \
+		GAsyncReadyCallback callback, \
+		gpointer user_data) \
+{ \
+	GTask *task; \
+	HexDocumentFindData *find_data = hex_document_find_data_new (); \
+ \
+	find_data->start = start; \
+	find_data->what = what; \
+	find_data->len = len; \
+	find_data->found_msg = found_msg; \
+	find_data->not_found_msg = not_found_msg; \
+ \
+	task = g_task_new (doc, cancellable, callback, user_data); \
+	g_task_set_return_on_cancel (task, TRUE); \
+	g_task_set_task_data (task, find_data, g_free); \
+	g_task_run_in_thread (task, FUNC_TO_CALL); \
+	g_object_unref (task);	/* _run_in_thread takes a ref */ \
+}
+
+FIND_ASYNC_TEMPLATE(hex_document_find_forward_async,
+		hex_document_find_forward_thread)
+
+/**
+ * hex_document_find_backward_full:
+ * @find_data: a #HexDocumentFindData structure
+ *
+ * Full version of [method@Hex.Document.find_backward] which allows for
+ * more flexibility than the above, which is only for a byte-by-byte exact
+ * match. However, it is less convenient to call since the caller must
+ * create and and free a #HexDocumentFindData structure manually.
+ *
+ * This method will block. For a non-blocking version, use
+ * [method@Hex.Document.find_backward_full_async].
+ *
+ * Returns: %TRUE if the search string contained in `find_data` was found by
+ *   the requested operation; %FALSE otherwise.
+ *
+ * Since: 4.2
+ */
+gboolean
+hex_document_find_backward_full (HexDocument *doc,
+		HexDocumentFindData *find_data)
 {
-	GTask *task;
-	HexDocumentFindData *find_data = g_new0 (HexDocumentFindData, 1);
+	gint64 pos = find_data->start;
+	
+	if (pos == 0)
+		return FALSE;
 
-	find_data->start = start;
-	find_data->what = what;
-	find_data->len = len;
-	find_data->found_msg = found_msg;
-	find_data->not_found_msg = not_found_msg;
+	do {
+		pos--;
+		if (hex_document_compare_data_full (doc, find_data, pos) == 0) {
+			find_data->offset = pos;
+			return TRUE;
+		}
+	} while (pos > 0);
 
-	task = g_task_new (doc, cancellable, callback, user_data);
-	g_task_set_return_on_cancel (task, TRUE);
-	g_task_set_task_data (task, find_data, g_free);
-	g_task_run_in_thread (task, hex_document_find_forward_thread);
+	return FALSE;
 }
 
 /**
@@ -1271,22 +1617,25 @@ gboolean
 hex_document_find_backward (HexDocument *doc, gint64 start, const char *what,
 						   size_t len, gint64 *offset)
 {
-	gint64 pos = start;
-	
-	if (pos == 0)
-		return FALSE;
+	gboolean retval;
+	HexDocumentFindData *find_data = hex_document_find_data_new ();
 
-	do {
-		pos--;
-		if (hex_document_compare_data (doc, what, pos, len) == 0) {
-			*offset = pos;
-			return TRUE;
-		}
-	} while (pos > 0);
+	find_data->start = start;
+	find_data->what = what;
+	find_data->len = len;
+	find_data->flags = HEX_SEARCH_NONE;
 
-	return FALSE;
+	retval = hex_document_find_backward_full (doc, find_data);
+	*offset = find_data->offset;
+
+	g_free (find_data);
+
+	return retval;
 }
 
+FIND_FULL_THREAD_TEMPLATE(hex_document_find_backward_full_thread,
+		hex_document_find_backward_full)
+
 static void
 hex_document_find_backward_thread (GTask *task,
 		gpointer source_object,
@@ -1303,6 +1652,20 @@ hex_document_find_backward_thread (GTask *task,
 	g_task_return_pointer (task, find_data, g_free);
 }
 
+/**
+ * hex_document_find_backward_full_async:
+ * @find_data: a #HexDocumentFindData structure
+ * @cancellable: (nullable): a #GCancellable
+ * @callback: (scope async): function to be called when the operation is
+ *   complete
+ *
+ * Non-blocking version of [method@Hex.Document.find_backward_full].
+ *
+ * Since: 4.2
+ */
+FIND_FULL_ASYNC_TEMPLATE(hex_document_find_backward_full_async,
+		hex_document_find_backward_full_thread)
+
 /**
  * hex_document_find_backward_async:
  * @doc: a [class@Hex.Document] object
@@ -1317,37 +1680,13 @@ hex_document_find_backward_thread (GTask *task,
  *   is not found
  * @callback: (scope async): function to be called when the operation is
  *   complete
- * 
+ *
  * Non-blocking version of [method@Hex.Document.find_backward]. This is the
  * function that should generally be used by a GUI client to find a string
  * backwards in a #HexDocument.
  */
-void
-hex_document_find_backward_async (HexDocument *doc,
-		gint64 start,
-		const char *what,
-		size_t len,
-		gint64 *offset,
-		const char *found_msg,
-		const char *not_found_msg,
-		GCancellable *cancellable,
-		GAsyncReadyCallback callback,
-		gpointer user_data)
-{
-	GTask *task;
-	HexDocumentFindData *find_data = g_new0 (HexDocumentFindData, 1);
-
-	find_data->start = start;
-	find_data->what = what;
-	find_data->len = len;
-	find_data->found_msg = found_msg;
-	find_data->not_found_msg = not_found_msg;
-
-	task = g_task_new (doc, cancellable, callback, user_data);
-	g_task_set_return_on_cancel (task, TRUE);
-	g_task_set_task_data (task, find_data, g_free);
-	g_task_run_in_thread (task, hex_document_find_backward_thread);
-}
+FIND_ASYNC_TEMPLATE(hex_document_find_backward_async,
+		hex_document_find_backward_thread)
 
 /**
  * hex_document_undo:
@@ -1460,9 +1799,6 @@ hex_document_real_redo(HexDocument *doc)
 				hex_document_set_byte(doc, cd->v_byte, cd->start, FALSE, FALSE);
 			else if(cd->rep_len == 0)
 				hex_document_set_byte(doc, cd->v_byte, cd->start, cd->insert, FALSE);
-#if 0
-				hex_document_delete_data(doc, cd->start, 1, FALSE);
-#endif
 			else
 				hex_document_set_byte(doc, cd->v_byte, cd->start, TRUE, FALSE);
 			cd->v_byte = c_val;
@@ -1549,6 +1885,8 @@ hex_document_get_undo_data (HexDocument *doc)
 HexBuffer *
 hex_document_get_buffer (HexDocument *doc)
 {
+	g_return_val_if_fail (HEX_IS_DOCUMENT (doc), NULL);
+
 	return doc->buffer;
 }
 
@@ -1564,5 +1902,31 @@ hex_document_get_buffer (HexDocument *doc)
 GFile *
 hex_document_get_file (HexDocument *doc)
 {
+	g_return_val_if_fail (HEX_IS_DOCUMENT (doc), NULL);
+
 	return doc->file;
 }
+
+/**
+ * hex_document_set_buffer:
+ * @doc: a [class@Hex.Document] object
+ * @buf: [iface@Hex.Buffer]
+ *
+ * Set the [iface@Hex.Buffer] connected with the #HexDocument.
+ *
+ * Returns: %TRUE if the operation was successful; %FALSE otherwise.
+ *
+ * Since: 4.2
+ */
+gboolean
+hex_document_set_buffer (HexDocument *doc, HexBuffer *buf)
+{
+	g_return_val_if_fail (HEX_IS_DOCUMENT (doc), FALSE);
+	g_return_val_if_fail (HEX_IS_BUFFER (buf), FALSE);
+
+	g_clear_object (&doc->buffer);
+	doc->buffer = buf;
+
+	g_object_notify_by_pspec (G_OBJECT(doc), properties[BUFFER]);
+	return TRUE;
+}
